# - Based on vesion 0.1.0.3-BETA of TorrentSpy from http://torrentspy.sf.net/
#   Original copyright unknown, SourceForge claims "MIT License"

package network::bittorrent;

use strict;
use Digest::SHA1 qw(sha1);

sub get_URL {
    require LWP::UserAgent;
    my $ua = LWP::UserAgent->new(
				 'env_proxy' => 1,
				 'agent'     => "MdkBitTorrent/0.1",
				 'timeout'   => 15,
				);
    return $ua->get($_[0]);
}

sub process_global_file_or_url {
    my $input = shift;
    my $result = {};
    my $body;
    if ($input =~ m!^(https?|ftp)://!) {
	my $res = get_URL($input);
	if (!$res->is_success) {
	    $result->{error} = "Problem connecting to URL $input\n: " . $res->status_line;
	} else {
	    $body = $res->content;
	}
    } else {
	my $TOR;
	if (!open($TOR, "<", $input)) {
	    $result->{error} = "Unable to read $input : $!\n";
	}
	binmode($TOR);
	read($TOR, $body, (-s $input));
	close($TOR);
    }
    if ($body) {
	my $t = bt_decode(\$body);
	my $info = $t->{info};
    
	$result->{hash} = bin2hex(sha1(substr($body, $t->{_info_start}, $t->{_info_length})));
	$result->{announce} = $t->{announce};
	$result->{files} = [];
	my $tsize = 0;
	if (defined($info->{files})) {
	    foreach my $f (@{$info->{files}}) {
		my %file_record = ('size' => $f->{length});
		
		$tsize += $f->{length};
		my $path = $f->{path};
		
		$file_record{name} = ref($path) eq 'ARRAY' ? $info->{name} . '/' . $path->[0] : $info->{name} . '/' . $path;
		push @{$result->{files}}, \%file_record;
	    }
	} else {
	    $tsize += $info->{length},
	    push @{$result->{files}}, { 'size' => $info->{length}, 'name' => $info->{name} };
	}
	$result->{total_size} = $tsize;
    }
    $result
}

sub get_torrent {
    my ($link, $login, $password) = @_;
    my @torrent;
    $login && $password and $link .= $login . $password;
    my $res = get_URL($link);
    if ($res->is_success) {
	@torrent = $res->content =~ m/<a href="(.*\.torrent)">/g;
    }
    @torrent;    
}

{
my %tracker_status_cache;
sub retrieve_tracker_info {
    my ($announce_url) = @_;
    
    if (exists $tracker_status_cache{$announce_url}) {
	return ($tracker_status_cache{$announce_url}, undef);
    }               
    
    my $scrape_url = get_scrape_url($announce_url);
    if (!defined $scrape_url) {
	return (undef,
		"Unable to determine URL to tracker information (scrape) server.");
    }

    my $res = get_URL($scrape_url);
    if (!$res->is_success) {
	my $error_reason = $res->status_; 
	if ($res->code == 404) {
	    $error_reason = "The tracker information server is not available. "
	      . "This tracker may not support providing information. "
		. "($error_reason)";
	}
	return (undef,
		"Unable to contact tracker information server at $scrape_url:"
		 . " $error_reason.");
    }
    
    my $tracker_status = get_tracker_status($res->content_ref);
    
    if (!defined $tracker_status) {
	return (undef,
                        "Error parsing results from tracker information server.");
	    }
    $tracker_status_cache{$announce_url} = $tracker_status;
    return ($tracker_status, undef);
}
}

sub get_tracker_status {
    my ($status_body) = @_;
    
    my $s;
    eval { $s = bt_decode($status_body) };
    if ($@) {
	print STDERR "Invalid tracker response $@";
	return undef;
    }
    if (!exists $s->{files}) {
	print STDERR "Tracker returned odd results (no files)\n";
	return undef;
    }
    
    my %results;
    foreach my $f (%{$s->{files}}) {
	my $v = $s->{files}{$f};
	my $fhash = bin2hex($f);
	my $seeds = $v->{complete} || "0";
	my $leeches = $v->{incomplete} || "0";
	if (exists $results{$fhash}) {
	    print STDERR "Tracker has hash $fhash multiple times\n";
	}
	$results{$fhash} = { 'complete' => $seeds, 'incomplete' => $leeches };
    }
    return \%results;
}

sub get_scrape_url {
    my $lk = shift;
    my $link = $lk;
    $link =~ s|/announce|/scrape|ig;
    $link eq $lk ? 'undef' : $link
}

sub bt_decode {
  my ($dataref) = @_;
  unless (ref($dataref) eq 'SCALAR') {
    die('Function bdecode takes a scalar ref!');
  } # unless
  my $p = 0;
  return bt_enc_parse_hash($dataref, \$p);
} # sub bdecode

sub bt_enc_parse_hash {
    my ($data, $p) = @_;
    my $c = substr($$data, $$p, 1);
    my $r;
    if ($c eq 'd') { # hash
	%$r = ();
	++$$p;
	while ($$p < length($$data) && substr($$data, $$p, 1) ne 'e') {
	    my $k = benc_parse_string($data, $p);
	    my $start = $$p;
	    $r->{'_' . $k . '_start'} = $$p if $k eq 'info';
	    my $v = bt_enc_parse_hash($data, $p);
	    $r->{'_' . $k . '_length'} = ($$p - $start)  if $k eq 'info';
	    $r->{$k} = $v;
	} 
	++$$p;
    } elsif ($c eq 'l') {
	@$r = \();
	++$$p;
	while (substr($$data, $$p, 1) ne 'e') {
	    push(@$r, bt_enc_parse_hash($data, $p));
	}
	++$$p;
    } elsif ($c eq 'i') {
	$r = 0;
	my $cc;
	++$$p;
	while (($cc = substr($$data, $$p, 1)) ne 'e') {
	    $r *= 10;
	    $r += int($cc);
	    ++$$p;
	}
	++$$p;
    } elsif ($c =~ /\d/) { # string
	$r = benc_parse_string($data, $p);
    } else {
	die("Unknown token '$c' at $p!");
    }
    return $r;
}
sub benc_parse_string {
    my ($data, $p) = @_;
    my $l = 0;
    my $c;
    my $s;
    while (($c = substr($$data, $$p, 1)) ne ':') {
	$l *= 10;
	$l += int($c);
	++$$p;
    }
    ++$$p;
    $s = substr($$data, $$p, $l);
    $$p += $l;
    return $s;
}
sub bin2hex {
  my ($d) = @_;
  $d =~ s/(.)/sprintf("%02x", ord($1))/egs;
  $d = lc($d);
  return $d;
}

1;
